@@ -17,19 +17,19 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <string.h>
 #include <gtk/gtk.h>
 #include "gailentry.h"
-#include "gailentryfactory.h"
 #include "gailcombo.h"
-#include "gailtexthelper.h"
+#include <libgail-util/gailmisc.h>
 
 static void       gail_entry_class_init            (GailEntryClass       *klass);
 static void       gail_entry_object_init           (GailEntry            *entry);
-static void	  gail_entry_real_init             (GailWidget           *widget,
-                                                    GtkWidget            *gtk_widget);
-static void	  gail_entry_real_notify_gtk	   (GObject		*obj,
-                                                    GParamSpec		*pspec);
-static void       gail_entry_finalize              (GObject            *object);
+static void	  gail_entry_real_initialize       (AtkObject            *obj,
+                                                    gpointer             data);
+static void	  gail_entry_real_notify_gtk	   (GObject		 *obj,
+                                                    GParamSpec		 *pspec);
+static void       gail_entry_finalize              (GObject              *object);
 
 static gint       gail_entry_get_index_in_parent   (AtkObject            *accessible);
 
@@ -133,8 +133,21 @@ static void	  _gail_entry_delete_text_cb	   (GtkEntry		 *entry,
                                                     gint		 arg1,
                                                     gint		 arg2);
 static void	  _gail_entry_changed_cb           (GtkEntry		 *entry);
-static gboolean    check_for_selection_change       (GailEntry           *entry,
-                                                     GtkEntry            *gail_entry);
+static gboolean   check_for_selection_change       (GailEntry            *entry,
+                                                    GtkEntry             *gail_entry);
+
+static void                  atk_action_interface_init   (AtkActionIface  *iface);
+
+static gboolean              gail_entry_do_action        (AtkAction       *action,
+                                                          gint            i);
+static gint                  gail_entry_get_n_actions    (AtkAction       *action);
+static G_CONST_RETURN gchar* gail_entry_get_description  (AtkAction       *action,
+                                                          gint            i);
+static G_CONST_RETURN gchar* gail_entry_action_get_name  (AtkAction       *action,
+                                                          gint            i);
+static gboolean              gail_entry_set_description  (AtkAction       *action,
+                                                          gint            i,
+                                                          const gchar     *desc);
 
 static GailWidgetClass *parent_class = NULL;
 
@@ -181,13 +194,21 @@ gail_entry_get_type (void)
         NULL
       };
 
+      static const GInterfaceInfo atk_action_info =
+      {
+        (GInterfaceInitFunc) atk_action_interface_init,
+        (GInterfaceFinalizeFunc) NULL,
+        NULL
+      };
+
       type = g_type_register_static (GAIL_TYPE_WIDGET,
                                      "GailEntry", &tinfo, 0);
       g_type_add_interface_static (type, ATK_TYPE_EDITABLE_TEXT,
                                    &atk_editable_text_info);
       g_type_add_interface_static (type, ATK_TYPE_TEXT,
                                    &atk_text_info);
-
+      g_type_add_interface_static (type, ATK_TYPE_ACTION,
+                                   &atk_action_info);
     }
   return type;
 }
@@ -201,47 +222,52 @@ gail_entry_class_init (GailEntryClass *klass)
 
   widget_class = (GailWidgetClass*)klass;
 
-  parent_class = g_type_class_ref (GAIL_TYPE_WIDGET);
+  parent_class = g_type_class_peek_parent (klass);
 
   gobject_class->finalize = gail_entry_finalize;
 
   class->ref_state_set = gail_entry_ref_state_set;
   class->get_index_in_parent = gail_entry_get_index_in_parent;
+  class->initialize = gail_entry_real_initialize;
 
-  widget_class->init = gail_entry_real_init;
   widget_class->notify_gtk = gail_entry_real_notify_gtk;
 }
 
 static void
 gail_entry_object_init (GailEntry *entry)
 {
-  entry->texthelper = NULL;
+  entry->textutil = NULL;
   entry->signal_name = NULL;
   entry->cursor_position = 0;
   entry->selection_bound = 0;
+  entry->activate_description = NULL;
 }
 
 static void
-gail_entry_real_init (GailWidget *widget, 
-                      GtkWidget  *gtk_widget)
+gail_entry_real_initialize (AtkObject *obj, 
+                            gpointer  data)
 {
   GtkEntry *entry;
+  GailEntry *gail_entry;
   const gchar *entry_text;
 
-  parent_class->init (widget, gtk_widget);
+  ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
-  GAIL_ENTRY (widget)->texthelper = gail_text_helper_new ();
+  gail_entry = GAIL_ENTRY (obj);
+  gail_entry->textutil = gail_text_util_new ();
   
-  entry = GTK_ENTRY (gtk_widget);
+  g_assert (GTK_IS_ENTRY (data));
+
+  entry = GTK_ENTRY (data);
   entry_text = gtk_entry_get_text (entry);
-  gail_text_helper_text_setup (GAIL_ENTRY(widget)->texthelper, entry_text);
+  gail_text_util_text_setup (gail_entry->textutil, entry_text);
 
   /* Set up signal callbacks */
-  g_signal_connect (G_OBJECT (gtk_widget), "insert-text",
+  g_signal_connect (data, "insert-text",
 	G_CALLBACK (_gail_entry_insert_text_cb), NULL);
-  g_signal_connect (G_OBJECT (gtk_widget), "delete-text",
+  g_signal_connect (data, "delete-text",
 	G_CALLBACK (_gail_entry_delete_text_cb), NULL);
-  g_signal_connect (G_OBJECT (gtk_widget), "changed",
+  g_signal_connect (data, "changed",
 	G_CALLBACK (_gail_entry_changed_cb), NULL);
 }
 
@@ -257,7 +283,7 @@ gail_entry_real_notify_gtk (GObject		*obj,
   if (strcmp (pspec->name, "cursor-position") == 0)
     {
       if (check_for_selection_change (entry, gtk_entry))
-        g_signal_emit_by_name (atk_obj, "selection_changed");
+        g_signal_emit_by_name (atk_obj, "text_selection_changed");
       /*
        * The entry cursor position has moved so generate the signal.
        */
@@ -267,7 +293,7 @@ gail_entry_real_notify_gtk (GObject		*obj,
   else if (strcmp (pspec->name, "selection-bound") == 0)
     {
       if (check_for_selection_change (entry, gtk_entry))
-        g_signal_emit_by_name (atk_obj, "selection_changed");
+        g_signal_emit_by_name (atk_obj, "text_selection_changed");
     }
   else if (strcmp (pspec->name, "editable") == 0)
     {
@@ -287,29 +313,28 @@ gail_entry_finalize (GObject            *object)
 {
   GailEntry *entry = GAIL_ENTRY (object);
 
-  g_object_unref (entry->texthelper);
+  g_object_unref (entry->textutil);
+  g_free (entry->activate_description);
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
-GtkAccessible* 
+AtkObject* 
 gail_entry_new (GtkWidget *widget)
 {
   GObject *object;
-  GtkAccessible *accessible;
+  AtkObject *accessible;
 
   g_return_val_if_fail (GTK_IS_ENTRY (widget), NULL);
 
   object = g_object_new (GAIL_TYPE_ENTRY, NULL);
 
-  g_return_val_if_fail (GTK_IS_ACCESSIBLE (object), NULL);
+  accessible = ATK_OBJECT (object);
+  atk_object_initialize (accessible, widget);
 
-  gail_widget_init (GAIL_WIDGET (object), widget);
-
-  accessible = GTK_ACCESSIBLE (object);
   if (GTK_ENTRY (widget)->visible)
-    ATK_OBJECT (accessible)->role = ATK_ROLE_TEXT;
+    accessible->role = ATK_ROLE_TEXT;
   else
-    ATK_OBJECT (accessible)->role = ATK_ROLE_PASSWORD_TEXT;
+    accessible->role = ATK_ROLE_PASSWORD_TEXT;
 
   return accessible;
 }
@@ -417,7 +442,7 @@ gail_entry_get_text_before_offset (AtkText	    *text,
   /* Get Entry */
   entry = GTK_ENTRY (widget);
 
-  return gail_text_helper_get_text (GAIL_ENTRY (text)->texthelper,
+  return gail_text_util_get_text (GAIL_ENTRY (text)->textutil,
                           gtk_entry_get_layout (entry), GAIL_BEFORE_OFFSET, 
                           boundary_type, offset, start_offset, end_offset);
 }
@@ -440,7 +465,7 @@ gail_entry_get_text_at_offset (AtkText          *text,
   /* Get Entry */
   entry = GTK_ENTRY (widget);
 
-  return gail_text_helper_get_text (GAIL_ENTRY (text)->texthelper,
+  return gail_text_util_get_text (GAIL_ENTRY (text)->textutil,
                             gtk_entry_get_layout (entry), GAIL_AT_OFFSET, 
                             boundary_type, offset, start_offset, end_offset);
 }
@@ -463,7 +488,7 @@ gail_entry_get_text_after_offset  (AtkText	    *text,
   /* Get Entry */
   entry = GTK_ENTRY (widget);
 
-  return gail_text_helper_get_text (GAIL_ENTRY (text)->texthelper,
+  return gail_text_util_get_text (GAIL_ENTRY (text)->textutil,
                            gtk_entry_get_layout (entry), GAIL_AFTER_OFFSET, 
                            boundary_type, offset, start_offset, end_offset);
 }
@@ -537,12 +562,12 @@ gail_entry_get_run_attributes (AtkText *text,
   dir = gtk_widget_get_direction (widget);
   if (dir == GTK_TEXT_DIR_RTL)
     {
-      at_set = gail_text_helper_add_attribute (at_set,
-                                               ATK_TEXT_ATTR_DIRECTION,
-       g_strdup (atk_attribute_get_value (ATK_TEXT_ATTR_DIRECTION, dir)));
+      at_set = gail_misc_add_attribute (at_set,
+                                        ATK_TEXT_ATTR_DIRECTION,
+       g_strdup (atk_text_attribute_get_value (ATK_TEXT_ATTR_DIRECTION, dir)));
     }
 
-  at_set = gail_text_helper_get_run_attributes (at_set,
+  at_set = gail_misc_layout_get_run_attributes (at_set,
                                                 gtk_entry_get_layout (entry),
                                                 entry->text,
                                                 offset,
@@ -565,16 +590,9 @@ gail_entry_get_default_attributes (AtkText *text)
 
   entry = GTK_ENTRY (widget);
 
-  at_set = gail_text_helper_add_attribute (at_set,
-                                           ATK_TEXT_ATTR_JUSTIFICATION,
-     g_strdup (atk_attribute_get_value (ATK_TEXT_ATTR_JUSTIFICATION, 0)));
-  at_set = gail_text_helper_add_attribute (at_set,
-                                           ATK_TEXT_ATTR_WRAP_MODE,
-              g_strdup (atk_attribute_get_value (ATK_TEXT_ATTR_WRAP_MODE, 0)));
-  at_set = gail_text_helper_get_default_attributes (at_set,
-                                                    gtk_entry_get_layout (entry)
-,
-                                                    widget);
+  at_set = gail_misc_get_default_attributes (at_set,
+                                             gtk_entry_get_layout (entry),
+                                             widget);
   return at_set;
 }
   
@@ -607,7 +625,7 @@ gail_entry_get_character_extents (AtkText *text,
     index += entry->preedit_length;
   pango_layout_index_to_pos (gtk_entry_get_layout(entry), index, &char_rect);
  
-  gail_text_helper_get_extents_from_pango_rectangle (widget, &char_rect, 
+  gail_misc_get_extents_from_pango_rectangle (widget, &char_rect, 
                         x_layout, y_layout, x, y, width, height, coords);
 } 
 
@@ -630,7 +648,7 @@ gail_entry_get_offset_at_point (AtkText *text,
   
   gtk_entry_get_layout_offsets (entry, &x_layout, &y_layout);
   
-  index = gail_text_helper_get_index_at_point_in_layout (widget, 
+  index = gail_misc_get_index_at_point_in_layout (widget, 
                gtk_entry_get_layout(entry), x_layout, y_layout, x, y, coords);
   if (index == -1)
     return -1;  
@@ -650,22 +668,6 @@ gail_entry_get_offset_at_point (AtkText *text,
     }
 }
 
-/* atkeditable.h */
-
-static void
-atk_editable_text_interface_init (AtkEditableTextIface *iface)
-{
-  g_return_if_fail (iface != NULL);
-
-  iface->set_text_contents = gail_entry_set_text_contents;
-  iface->insert_text = gail_entry_insert_text;
-  iface->copy_text = gail_entry_copy_text;
-  iface->cut_text = gail_entry_cut_text;
-  iface->delete_text = gail_entry_delete_text;
-  iface->paste_text = gail_entry_paste_text;
-  iface->set_run_attributes = NULL;;
-}
-
 static gint
 gail_entry_get_n_selections (AtkText              *text)
 {
@@ -817,12 +819,27 @@ gail_entry_set_selection (AtkText *text,
     return FALSE;
 }
 
+static void
+atk_editable_text_interface_init (AtkEditableTextIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+
+  iface->set_text_contents = gail_entry_set_text_contents;
+  iface->insert_text = gail_entry_insert_text;
+  iface->copy_text = gail_entry_copy_text;
+  iface->cut_text = gail_entry_cut_text;
+  iface->delete_text = gail_entry_delete_text;
+  iface->paste_text = gail_entry_paste_text;
+  iface->set_run_attributes = NULL;
+}
+
 static void
 gail_entry_set_text_contents (AtkEditableText *text,
                               const gchar     *string)
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -830,6 +847,9 @@ gail_entry_set_text_contents (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
   gtk_entry_set_text (entry, string);
 }
@@ -842,6 +862,7 @@ gail_entry_insert_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -849,8 +870,11 @@ gail_entry_insert_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
-  gtk_editable_insert_text (GTK_EDITABLE (entry), string, length, position);
+  gtk_editable_insert_text (editable, string, length, position);
 }
 
 static void
@@ -860,6 +884,7 @@ gail_entry_copy_text   (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
   gchar *str;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -868,8 +893,9 @@ gail_entry_copy_text   (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
-  str = gtk_editable_get_chars (GTK_EDITABLE (entry), start_pos, end_pos);
-  gtk_clipboard_set_text(gtk_clipboard_get(GDK_NONE), str, -1);
+  editable = GTK_EDITABLE (entry);
+  str = gtk_editable_get_chars (editable, start_pos, end_pos);
+  gtk_clipboard_set_text (gtk_clipboard_get (GDK_NONE), str, -1);
 }
 
 static void
@@ -879,6 +905,7 @@ gail_entry_cut_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
   gchar *str;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -887,9 +914,12 @@ gail_entry_cut_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
-  str = gtk_editable_get_chars (GTK_EDITABLE (entry), start_pos, end_pos);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
+  str = gtk_editable_get_chars (editable, start_pos, end_pos);
   gtk_clipboard_set_text (gtk_clipboard_get (GDK_NONE), str, -1);
-  gtk_editable_delete_text (GTK_EDITABLE (entry), start_pos, end_pos);
+  gtk_editable_delete_text (editable, start_pos, end_pos);
 }
 
 static void
@@ -899,6 +929,7 @@ gail_entry_delete_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -906,8 +937,11 @@ gail_entry_delete_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
-  gtk_editable_delete_text (GTK_EDITABLE (entry), start_pos, end_pos);
+  gtk_editable_delete_text (editable, start_pos, end_pos);
 }
 
 static void
@@ -915,6 +949,7 @@ gail_entry_paste_text (AtkEditableText *text,
                        gint            position)
 {
   GtkWidget *widget;
+  GtkEditable *editable;
   GailEntryPaste paste_struct;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -922,6 +957,9 @@ gail_entry_paste_text (AtkEditableText *text,
     /* State is defunct */
     return;
 
+  editable = GTK_EDITABLE (widget);
+  if (!gtk_editable_get_editable (editable))
+    return;
   paste_struct.entry = GTK_ENTRY (widget);
   paste_struct.position = position;
 
@@ -979,6 +1017,7 @@ gail_entry_get_character_at_offset (AtkText	         *text,
   GtkEntry *entry;
   gchar *string;
   gchar *index;
+  gunichar unichar;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -988,9 +1027,10 @@ gail_entry_get_character_at_offset (AtkText	         *text,
   entry = GTK_ENTRY (widget);
   string = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
   index = g_utf8_offset_to_pointer (string, offset);
-  g_free(string);
 
-  return g_utf8_get_char(index);
+  unichar = g_utf8_get_char(index);
+  g_free(string);
+  return unichar;
 }
 
 /* Note arg1 returns the start of the delete range, arg2 returns the
@@ -1028,8 +1068,7 @@ _gail_entry_changed_cb (GtkEntry *entry)
   /*
    * Update the text in the GtkTextBuffer
    */
-  gail_text_helper_text_setup (gail_entry->texthelper, 
-                               gtk_entry_get_text(entry));
+  gail_text_util_text_setup (gail_entry->textutil, gtk_entry_get_text(entry));
 }
 
 static gboolean 
@@ -1049,7 +1088,7 @@ check_for_selection_change (GailEntry   *entry,
          * for notification of selection_bound or current_pos.
          * The values of current_pos  selection_bound may be the same 
          * for both notifications and we only want to generate one
-         * selection_changed signal.
+         * text_selection_changed signal.
          */
         ret_val = TRUE;
     }
@@ -1063,3 +1102,117 @@ check_for_selection_change (GailEntry   *entry,
 
   return ret_val;
 }
+
+static void
+atk_action_interface_init (AtkActionIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+
+  iface->do_action = gail_entry_do_action;
+  iface->get_n_actions = gail_entry_get_n_actions;
+  iface->get_description = gail_entry_get_description;
+  iface->get_name = gail_entry_action_get_name;
+  iface->set_description = gail_entry_set_description;
+}
+
+static gboolean
+gail_entry_do_action (AtkAction *action,
+                      gint      i)
+{
+  GtkWidget *widget;
+  gboolean return_value = TRUE;
+
+  widget = GTK_ACCESSIBLE (action)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return FALSE;
+
+  if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    return FALSE;
+
+  switch (i)
+    {
+    case 0:
+      gtk_widget_activate (widget);
+      break;
+    default:
+      return_value = FALSE;
+      break;
+    }
+  return return_value; 
+}
+
+static gint
+gail_entry_get_n_actions (AtkAction *action)
+{
+  return 1;
+}
+
+static G_CONST_RETURN gchar*
+gail_entry_get_description (AtkAction *action,
+                            gint      i)
+{
+  GailEntry *entry;
+  G_CONST_RETURN gchar *return_value;
+
+  entry = GAIL_ENTRY (action);
+  switch (i)
+    {
+    case 0:
+      return_value = entry->activate_description;
+      break;
+    default:
+      return_value = NULL;
+      break;
+    }
+  return return_value; 
+}
+
+static G_CONST_RETURN gchar*
+gail_entry_action_get_name (AtkAction *action,
+                            gint      i)
+{
+  G_CONST_RETURN gchar *return_value;
+
+  switch (i)
+    {
+    case 0:
+      return_value = "activate";
+      break;
+    default:
+      return_value = NULL;
+      break;
+  }
+  return return_value; 
+}
+
+static gboolean
+gail_entry_set_description (AtkAction      *action,
+                            gint           i,
+                            const gchar    *desc)
+{
+  GailEntry *entry;
+  gchar **value;
+
+  entry = GAIL_ENTRY (action);
+  switch (i)
+    {
+    case 0:
+      value = &entry->activate_description;
+      break;
+    default:
+      value = NULL;
+      break;
+    }
+
+  if (value)
+    {
+      g_free (*value);
+      *value = g_strdup (desc);
+      return TRUE;
+    }
+  else
+    return FALSE;
+}
