@@ -50,7 +50,7 @@ struct _GailItemClass
   GailContainerClass parent_class;
 };
 
-GtkAccessible *gail_item_new(GtkWidget *widget);
+AtkObject* gail_item_new (GtkWidget *widget);
 
 #ifdef __cplusplus
 }
