@@ -20,7 +20,7 @@
 #include <string.h>
 #include <gtk/gtk.h>
 #include "gaillabel.h"
-#include <util/gailmisc.h>
+#include <libgail-util/gailmisc.h>
 
 static void       gail_label_class_init            (GailLabelClass    *klass);
 static void	  gail_label_real_initialize	   (AtkObject 	      *obj,
@@ -444,7 +444,7 @@ gail_label_get_text (AtkText *text,
   else
   {
     return gail_text_util_get_substring (GAIL_LABEL(text)->textutil, 
-                                         &start_pos, &end_pos);
+                                         start_pos, end_pos);
   }
 }
 
@@ -621,7 +621,7 @@ gail_label_get_selection (AtkText *text,
         return 0;
       else
         return gail_text_util_get_substring (GAIL_LABEL (text)->textutil, 
-                                             start_pos, end_pos);
+                                             *start_pos, *end_pos);
     }
   else 
     return NULL;
