@@ -139,7 +139,7 @@ gail_menu_item_new (GtkWidget *widget)
 {
   GObject *object;
   AtkObject *accessible;
-
+  
   g_return_val_if_fail (GTK_IS_MENU_ITEM (widget), NULL);
 
   object = g_object_new (GAIL_TYPE_MENU_ITEM, NULL);
@@ -156,6 +156,8 @@ gail_menu_item_new (GtkWidget *widget)
       accessible->role = ATK_ROLE_CHECK_MENU_ITEM;
   else if (GTK_IS_SEPARATOR_MENU_ITEM (widget))
     accessible->role = ATK_ROLE_SEPARATOR;
+  if (gtk_menu_item_get_submenu (GTK_MENU_ITEM(widget)))
+    accessible->role = ATK_ROLE_MENU;
   else
     accessible->role = ATK_ROLE_MENU_ITEM;
 
@@ -254,6 +256,9 @@ gail_menu_item_do_action (AtkAction *action,
         /* State is defunct */
         return FALSE;
 
+      if (!GTK_WIDGET_SENSITIVE (item) || !GTK_WIDGET_VISIBLE (item))
+        return FALSE;
+
       gtk_menu_shell_select_item (GTK_MENU_SHELL (item->parent), item);
       /*
        * This is what is called when <Return> is pressed for a menu item
@@ -338,8 +343,9 @@ gail_menu_item_get_keybinding (AtkAction *action,
           child = gtk_bin_get_child (GTK_BIN (temp_item));
           if (child == NULL)
             {
-              /* Probably a tear off menu item */
-              g_return_val_if_fail (GTK_IS_TEAROFF_MENU_ITEM (temp_item), NULL);
+              /* Possibly a tear off menu item; it could also be a menu 
+               * separator generated by gtk_item_factory_create_items()
+               */
               return NULL;
             }
           parent = gtk_widget_get_parent (temp_item);
