@@ -52,7 +52,7 @@ struct _GailPixmapClass
   GailWidgetClass parent_class;
 };
 
-GtkAccessible *gail_pixmap_new(GtkWidget *widget);
+AtkObject* gail_pixmap_new (GtkWidget *widget);
 
 #ifdef __cplusplus
 }
