@@ -34,6 +34,8 @@ static gboolean              gail_window_real_focus_gtk (GtkWidget     *widget,
                                                          GdkEventFocus *event);
 
 static AtkStateSet*          gail_window_ref_state_set  (AtkObject     *accessible);
+static void                  gail_window_real_notify_gtk (GObject      *obj,
+                                                          GParamSpec   *pspec);
 
 static gboolean              gail_window_window_state_event_gtk
                                                         (GtkWidget           *widget,
@@ -77,6 +79,7 @@ gail_window_class_init (GailWindowClass *klass)
 
   widget_class = (GailWidgetClass*)klass;
   widget_class->focus_gtk = gail_window_real_focus_gtk;
+  widget_class->notify_gtk = gail_window_real_notify_gtk;
 
   parent_class = g_type_class_peek_parent (klass);
 
@@ -263,6 +266,28 @@ gail_window_ref_state_set (AtkObject *accessible)
   return state_set;
 }
 
+static void
+gail_window_real_notify_gtk (GObject		*obj,
+                             GParamSpec		*pspec)
+{
+  GtkWidget *widget = GTK_WIDGET (obj);
+  AtkObject* atk_obj = gtk_widget_get_accessible (widget);
+
+  if (strcmp (pspec->name, "title") == 0)
+    {
+      if (atk_obj->name == NULL)
+        {
+        /*
+         * The title has changed so notify a change in accessible-name
+         */
+          g_object_notify (G_OBJECT (atk_obj), "accessible-name");
+	}
+      g_signal_emit_by_name (atk_obj, "visible_data_changed");
+    }
+  else
+    GAIL_WIDGET_CLASS (parent_class)->notify_gtk (obj, pspec);
+}
+
 static gboolean
 gail_window_window_state_event_gtk (GtkWidget           *widget,
                                     GdkEventWindowState *event)
